/*
 * Netris -- A free networked version of T*tris
 * Copyright (C) 1994-1996,1999  Mark H. Weaver <mhw@netris.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "netris.h"
#include <stdlib.h>

#include "board.h"

static const char shapes[7][4][4][4] = {
	/*
	 * 4 rotations of 4x4 pixels per shape
	 * high nibble signifies joinage (left, right, top, bottom, from MSB)
	 * low nibble identifies block type (typically 2..8)
	 */
	
	// J
	{ { {   0,    0,    0, 0000},
	    {0x47, 0xC7, 0x97, 0000},
	    {   0,    0, 0x27, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0, 0x17,    0, 0000},
	    {   0, 0x37,    0, 0000},
	    {0x47, 0xA7,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {0x17,    0,    0, 0000},
	    {0x67, 0xC7, 0x87, 0000},
	    {   0,    0,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0, 0x57, 0x87, 0000},
	    {   0, 0x37,    0, 0000},
	    {   0, 0x27,    0, 0000},
	    {0000, 0000, 0000, 0000} } },
	
	// L
	{ { {   0,    0,    0, 0000},
	    {0x53, 0xC3, 0x83, 0000},
	    {0x23,    0,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {0x43, 0x93,    0, 0000},
	    {   0, 0x33,    0, 0000},
	    {   0, 0x23,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0,    0, 0x13, 0000},
	    {0x43, 0xC3, 0xA3, 0000},
	    {   0,    0,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0, 0x13,    0, 0000},
	    {   0, 0x33,    0, 0000},
	    {   0, 0x63, 0x83, 0000},
	    {0000, 0000, 0000, 0000} } },
	
	// T
	{ { {   0,    0,    0, 0000},
	    {0x48, 0xD8, 0x88, 0000},
	    {   0, 0x28,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0, 0x18,    0, 0000},
	    {0x48, 0xB8,    0, 0000},
	    {   0, 0x28,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0, 0x18,    0, 0000},
	    {0x48, 0xE8, 0x88, 0000},
	    {   0,    0,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0, 0x18,    0, 0000},
	    {   0, 0x78, 0x88, 0000},
	    {   0, 0x28,    0, 0000},
	    {0000, 0000, 0000, 0000} } },
	
	// S
	{ { {   0,    0,    0, 0000},
	    {   0, 0x52, 0x82, 0000},
	    {0x42, 0xA2,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {0x12,    0,    0, 0000},
	    {0x62, 0x92,    0, 0000},
	    {   0, 0x22,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0,    0,    0, 0000},
	    {   0, 0x52, 0x82, 0000},
	    {0x42, 0xA2,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {0x12,    0,    0, 0000},
	    {0x62, 0x92,    0, 0000},
	    {   0, 0x22,    0, 0000},
	    {0000, 0000, 0000, 0000} } },
	
	// Z
	{ { {   0,    0,    0, 0000},
	    {0x46, 0x96,    0, 0000},
	    {   0, 0x66, 0x86, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0, 0x16,    0, 0000},
	    {0x56, 0xA6,    0, 0000},
	    {0x26,    0,    0, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0,    0,    0, 0000},
	    {0x46, 0x96,    0, 0000},
	    {   0, 0x66, 0x86, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0, 0x16,    0, 0000},
	    {0x56, 0xA6,    0, 0000},
	    {0x26,    0,    0, 0000},
	    {0000, 0000, 0000, 0000} } },
	
	// I
	{ { {   0,    0,    0,    0},
	    {0x44, 0xC4, 0xC4, 0x84},
	    {   0,    0,    0,    0},
	    {   0,    0,    0,    0} },
	  { {   0, 0x14,    0,    0},
	    {   0, 0x34,    0,    0},
	    {   0, 0x34,    0,    0},
	    {   0, 0x24,    0,    0} },
	  { {   0,    0,    0,    0},
	    {0x44, 0xC4, 0xC4, 0x84},
	    {   0,    0,    0,    0},
	    {   0,    0,    0,    0} },
	  { {   0, 0x14,    0,    0},
	    {   0, 0x34,    0,    0},
	    {   0, 0x34,    0,    0},
	    {   0, 0x24,    0,    0} } },
	
	// O
	{ { {   0,    0,    0, 0000},
	    {   0, 0x55, 0x95, 0000},
	    {   0, 0x65, 0xA5, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0,    0,    0, 0000},
	    {   0, 0x55, 0x95, 0000},
	    {   0, 0x65, 0xA5, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0,    0,    0, 0000},
	    {   0, 0x55, 0x95, 0000},
	    {   0, 0x65, 0xA5, 0000},
	    {0000, 0000, 0000, 0000} },
	  { {   0,    0,    0, 0000},
	    {   0, 0x55, 0x95, 0000},
	    {   0, 0x65, 0xA5, 0000},
	    {0000, 0000, 0000, 0000} } }
};

int shape_iterate(char s, int scr, int y, int x, ShapeDrawFunc func)
{ //Draw a certain shape using <ShapeDrawFunc>
	int i, j, result;
	char type, rotation;

	type = s / 4;
	rotation = s & 3;
	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			if (shapes[type][rotation][i][j])
				if (result = func(scr, y-i, x+j, shapes[type][rotation][i][j]))
					return result;
	return 0;
}


float stdOptions[7] = {1, 1, 1, 1, 1, 1, 1}; //stdOptions

char ChooseOption(float options[7])
{ //Return a random piece with given piece weight
	int i;
	float total = 0, val;

	for (i = 0; i < 7; i++) total += options[i];
	val = Random(0, 32767) / 32768.0 * total;
	for (i = 0; i < 7; i++) if ((val -= options[i]) < 0)
		return i << 2;
	return 0;
}


static unsigned char board[MAX_SCREENS][MAX_BOARD_HEIGHT][MAX_BOARD_WIDTH];
static unsigned char oldBoard[MAX_SCREENS][MAX_BOARD_HEIGHT][MAX_BOARD_WIDTH];
static unsigned int changed[MAX_SCREENS][MAX_BOARD_HEIGHT];
static int shadowy;

void player_empty(int scr)
{ //Empty the whole field (all blocks BT_none)
	int y, x;

	for (y = Players[scr].boardHeight - 1; y >= 0; --y)
		for (x = 0; x < Players[scr].boardWidth; ++x) {
			oldBoard[scr][y][x] = board[scr][y][x] = BT_none;
		}
}

unsigned char block_get(int scr, int y, int x)
{ //Returns the block on field at position (x,y)
	if (y < 0 || x < 0 || x >= Players[scr].boardWidth)
		return BT_wall;
	else if (y >= Players[scr].boardHeight)
		return BT_none;
	else
		return board[scr][y][x];
}

static void block_set(int scr, int y, int x, unsigned char type)
{
	if (y >= 0 && y < Players[scr].boardHeight
	 && x >= 0 && x < Players[scr].boardWidth) {
		board[scr][y][x] = type;
		changed[scr][y] |= 1 << x;
	}
}

int player_draw(int scr)
{ //draw changes to screen
	int y, x, any = 0;
	unsigned int c;

	for (y = Players[scr].boardVisible - 1; y >= 0; y--)
		if ((c = changed[scr][y])) { // line changed
			for (x = 0; c; (c >>= 1), x++)
				if (c & 1 && board[scr][y][x] != oldBoard[scr][y][x]) {
					block_draw_window(scr, y, x, board[scr][y][x]);
					oldBoard[scr][y][x] = board[scr][y][x];
				}
			changed[scr][y] = 0; // reset
			any = 1;
		} //changed row
	return any;
}

int block_iter_set_status(int scr, int y, int x, unsigned char type)
{
	block_draw_status(y, x, type);
	return 0;
}

static int block_iter_shadow(int scr, int y, int x, unsigned char type)
{ //draw shadow
	block_set(scr, y, x, BT_shadow);
	return 0;
}

static int block_iter_set(int scr, int y, int x, unsigned char type)
{
	block_set(scr, y, x, type);
	return 0;
}

void shape_draw(char shape, int scr, int y, int x, int shadow)
{ //put shape on field
	if (shadow) {
		for (shadowy = y - 1; shadowy >= 0; shadowy--)
			if (!shape_get(shape, scr, shadowy, x))
				break;
		shape_iterate(shape, scr, shadowy + 1, x, block_iter_shadow);
	} //draw shadow
	shape_iterate(shape, scr, y, x, block_iter_set);
}

static int block_iter_erase(int scr, int y, int x, unsigned char type)
{
	block_set(scr, y, x, BT_none);
	return 0;
}

void shape_erase(char shape, int scr, int y, int x, int shadow)
{ //remove block from field
	shape_iterate(shape, scr, y, x, block_iter_erase);
	if (shadow && scr == me) // draw shadow
		shape_iterate(shape, scr, shadowy + 1, x, block_iter_erase);
}

static int block_iter_get(int scr, int y, int x, unsigned char type)
{
	return block_get(scr, y, x) > BT_none;
}

int shape_get(char shape, int scr, int y, int x)
{ //check if there's nothing in the way
	return !shape_iterate(shape, scr, y, x, block_iter_get);
}

static int block_iter_visible(int scr, int y, int x, unsigned char type)
{
	return (y >= 0 && y < Players[scr].boardVisible &&
			x >= 0 && x < Players[scr].boardWidth);
}
int shape_visible(char shape, int scr, int y, int x)
{
	return shape_iterate(shape, scr, y, x, block_iter_visible);
}

int player_move(int scr, int deltaY, int deltaX)
{
	int result;

	shape_erase(Players[scr].curShape, scr,
		Players[scr].curY, Players[scr].curX, 1);
	result = shape_get(Players[scr].curShape, scr, Players[scr].curY + deltaY,
	                   Players[scr].curX + deltaX);
	if (result) {
		Players[scr].curY += deltaY;
		Players[scr].curX += deltaX;
	}
	shape_draw(Players[scr].curShape, scr, Players[scr].curY, Players[scr].curX,
		scr == me);
	return result;
}

int player_rotate(int scr, int dir)
{
	char newshape;
	int result;

	shape_erase(Players[scr].curShape, scr, Players[scr].curY,
		Players[scr].curX, 1);
	/* (inc|dec)rement only 3 least significant bits which indicate rotation */
	newshape = (Players[scr].curShape & 252) + (((Players[scr].curShape & 3) + dir) & 3);
	result = shape_get(newshape, scr, Players[scr].curY, Players[scr].curX);
	if (!result) {
		// move if it doesn't fit anymore
		short int slideX;
		for (slideX = 0; slideX < 2; slideX = -slideX) {
			// slide left and right
			if (slideX >= 0) slideX++; // slide more
			if (result = shape_get(newshape, scr, Players[scr].curY,
				Players[scr].curX+slideX)) break;
		}
		if (result) Players[scr].curX += slideX;
	}
	if (result) Players[scr].curShape = newshape;
	shape_draw(Players[scr].curShape, scr,
		Players[scr].curY, Players[scr].curX, scr == me);
	return result;
}

int player_drop(int scr)
{
	int count = 0;

	shape_erase(Players[scr].curShape, scr,
		Players[scr].curY, Players[scr].curX, 1);
	while (shape_get(Players[scr].curShape, scr,
	       Players[scr].curY - 1, Players[scr].curX)) {
		Players[scr].curY--;
		count++;
	}
	shape_draw(Players[scr].curShape, scr,
		Players[scr].curY, Players[scr].curX, 0);
	return count;
}

static int block_free(int scr, int x, int y, unsigned char z)
{ //Check if blocks are empty below block (x,y) and sticking to (x,y) mask <z>
	unsigned char curblock;

	if (y == 0) return 0; // at bottom
	curblock = block_get(scr, y, x) & z;
	if (curblock & 0x10 && !block_free(scr, x, y-1, z & 0xD0)) return 0;
	if (curblock & 0x20 && !block_free(scr, x, y+1, z & 0xE0)) return 0;
	if (curblock & 0x40 && !block_free(scr, x+1, y, z & 0x70)) return 0;
	if (curblock & 0x80 && !block_free(scr, x-1, y, z & 0xB0)) return 0;
	if ((z = block_get(scr, y-1, x)) & 0x20) return 1; // stuck to block below
	if (z > BT_none) return 0; // some other piece below
	return 1; // nothing below
}

static int block_down(int scr, int x, int y, unsigned char z)
{ //Drop down block (x,y) and those sticking to it mask <z>
	if (block_get(scr, y, x) & z & 0x10) block_down(scr, x, y-1, z & 0xD0);
	if (block_get(scr, y, x) & z & 0x20) block_down(scr, x, y+1, z & 0xE0);
	if (block_get(scr, y, x) & z & 0x40) block_down(scr, x+1, y, z & 0x70);
	if (block_get(scr, y, x) & z & 0x80) block_down(scr, x-1, y, z & 0xB0);
	block_set(scr, y-1, x, block_get(scr, y, x));
	block_set(scr, y, x, BT_none);
}

int player_down(int scr)
{ //Drop any free blocks on field
	int xloop, x, x2, y, fallen = 0;
	unsigned char z;

	if (!Game.gravity) return 0;
	for (y = Players[scr].boardHeight - 1; y > 0; y--)
		for (x = 0; x < Players[scr].boardWidth; x++) {
			if ((z = block_get(scr, y, x)) > BT_none && (z & 0xA0) == 0) {
				// block present which doesn't stick left/up => topleft block
				if (block_free(scr, x, y, 0xF0)) {
					block_down(scr, x, y, 0xF0); // move blocks down
					fallen++;
				}
			} //block present
		}
	return fallen;
}

static int player_linecheck(int scr, int y)
{ //return 0 if any blocks present on line <y>
	int x;

	for (x = 0; x < Players[scr].boardWidth; x++)
		if (block_get(scr, y, x) <= BT_none)
			return 0;
	return 1;
}

static void player_linecopy(int scr, int from, int to)
{ //move blocks on line <from> to line <to>
	int x;

	if (from != to)
		for (x = 0; x < Players[scr].boardWidth; ++x)
			block_set(scr, to, x, block_get(scr, from, x));
}

int player_lineclear(int scr)
{ //remove full lines, return lines cleared
	int from, to, x, linescleared = 0;

	do {
	from = to = 0;
	while (to < Players[scr].boardHeight) {
		while (player_linecheck(scr, from)) {
			from++; // skip
			for (x = 0; x<Players[scr].boardWidth; x++) {
				// don't stick blocks to line which we'll remove
				block_set(scr, from, x, block_get(scr, from, x) & 0xEF);
				if (from > 1)
					block_set(scr, from-2, x, block_get(scr, from-2, x) & 0xDF);
			}
		}
		player_linecopy(scr, from++, to++);
	}
	linescleared += from - to;
	} while (player_down(scr));
	return linescleared;
}

void player_lineadd(int scr, int color, int count, int column)
{ //add <count> junklines with hole at <column> to <scr> by team <color>
	int y, x;

	shape_erase(Players[scr].curShape, scr,
		Players[scr].curY, Players[scr].curX, 1);
	for (y = Players[scr].boardHeight - count - 1; y >= 0; --y)
		player_linecopy(scr, y, y + count);
	for (y = 0; y < count; ++y)
		for (x = 0; x < Players[scr].boardWidth; ++x)
			block_set(scr, y, x, x == column ? BT_none : color + 1
				+ 0x40 * (x != column-1 && x < Players[scr].boardWidth-1)
				+ 0x80 * (x != column+1 && x > 0));
	Players[scr].curY += count; // move piece up..
	for (y = 0; y < count; ++y)
		if (shape_get(Players[scr].curShape, scr, Players[scr].curY - 1, Players[scr].curX))
			Players[scr].curY--; // ...and down again as far as possible
		else break;
	shape_draw(Players[scr].curShape, scr, Players[scr].curY, Players[scr].curX,
		scr == me);
}

